.workspace-dashboard__page-heading{display:grid;min-height:2.75rem}@media (min-width: 768px){.workspace-dashboard__page-heading{display:flex;flex-direction:row;align-items:center}}.workspace-dashboard__header-actions{display:flex;width:100%;max-width:100%;flex-direction:column;align-items:center;gap:.5rem}@media (min-width: 768px){.workspace-dashboard__header-actions{display:inline-flex;width:auto;flex-direction:row}}.workspace-dashboard__date-select{min-width:0px}.workspace-dashboard__grid{display:grid;grid-template-columns:repeat(1,minmax(0,1fr));align-items:flex-start;gap:1rem}@media (min-width: 1280px){.workspace-dashboard__grid{grid-template-columns:repeat(2,minmax(0,1fr))}}.workspace-dashboard__side{display:grid;grid-template-columns:repeat(1,minmax(0,1fr));gap:1rem}.workspace-dashboard__subflows-toggle{width:100%;padding-right:.5rem}@media (min-width: 768px){.workspace-dashboard__subflows-toggle{width:auto}}
