.runs__scatter-plot{height:275px!important}.runs__chameleon-link{cursor:pointer;font-weight:600;color:var(--p-color-text-link)}.runs__chameleon-link:hover{text-decoration-line:underline}.runs__subflows-toggle{width:100%;padding-right:.5rem}@media (min-width: 768px){.runs__subflows-toggle{width:auto}}
