.context-sidebar__logo-link{border-radius:.375rem;outline:2px solid transparent;outline-offset:2px}.context-sidebar__logo-link:focus{--tw-ring-offset-shadow: var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color);--tw-ring-shadow: var(--tw-ring-inset) 0 0 0 calc(var(--p-spacing-focus-ring) + var(--tw-ring-offset-width)) var(--tw-ring-color);box-shadow:var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow, 0 0 #0000);--tw-ring-color: var(--p-color-focus-ring) }.context-sidebar__logo-link:focus:not(:focus-visible){--tw-ring-color: transparent }.context-sidebar__logo-icon{height:42px!important;width:42px!important}.context-sidebar__upgade-button{margin-left:auto}.body-scrolling-disabled{overflow:hidden}.app-router-view{display:flex;flex-direction:column;overflow:auto;background-repeat:no-repeat;--prefect-scroll-margin: 5rem;height:100vh;background-image:url(/PREFECT_UI_SERVE_BASE_REPLACE_PLACEHOLDER/decorative_iso-pixel-grid_light.svg);background-attachment:fixed;background-position:bottom -140px left -140px}.dark .app-router-view{background-image:url(/PREFECT_UI_SERVE_BASE_REPLACE_PLACEHOLDER/decorative_iso-pixel-grid_dark.svg)}.app-router-view__prefect-icon{height:1.75rem;width:1.75rem}.app-router-view__mobile-menu{height:auto;padding-top:.75rem;padding-bottom:.75rem}.app-router-view__sidebar{top:54px;background-color:var(--p-color-bg-floating)}@media (min-width: 1024px){.app-router-view__sidebar{top:0;background-color:transparent}}.app-router-view__sidebar .p-context-sidebar__header{display:none}@media (min-width: 1024px){.app-router-view__sidebar .p-context-sidebar__header{display:block}}.app-router-view__view{flex:1 0 1px;height:100%}.app-router-view__view--public{display:flex;align-items:center;justify-content:center;grid-column:1 / -1}@media (min-width: 1024px){.app-router-view{--prefect-scroll-margin: .5rem;display:grid;grid-template-columns:max-content minmax(0,1fr)}}.app-router-view-fade-enter-active,.app-router-view-fade-leave-active{transition:opacity .25s ease}.app-router-view-fade-enter-from,.app-router-view-fade-leave-to{opacity:0}
