function r(n,e){if(e.length<n)throw new TypeError(n+" argument required, but only "+e.length+" present")}var o=1e3;function t(n){return r(1,arguments),n*o}export{t as s};
//# sourceMappingURL=index-B4HswuBc.js.map
