import{d as e,c as o,o as a,m as t,dv as r}from"./index-g6FNXcTE.js";import{u as s}from"./usePageTitle-DEr56mEx.js";const _=e({__name:"404",setup(n){return s("Page Not Found"),(c,m)=>(a(),o(t(r)))}});export{_ as default};
//# sourceMappingURL=404-CXo2ZBur.js.map
